from typing import Any, Dict
import json

import pandas as pd
from enums.DatabaseTypeEnum import DatabaseTypeEnum
from enums.ProcessStatusEnum import ProcessStatusEnum
from components.Utils import select_multiple, generateModel
from components.DatabaseOperation.DatabaseExtraction import get_iterator, get_steps
from components.DatabaseOperation.DatabaseLoad import save_from_dataframe
from components.DatabaseOperation.DatabaseTransformation import delete_table
from components.Control import get_tasks_from_control, update_new_process
from components.S3Route import load_obj_to_s3
from components.Xcom import delete_all_xcom_tasks, delete_task_instances
from enums.OperationTypeEnum import OperationTypeEnum

from airflow.utils.task_group import TaskGroup
from airflow.operators.python import PythonOperator
from airflow.models import Variable
from airflow.decorators import task
from airflow.exceptions import AirflowSkipException
import logging

logger = logging.getLogger()


def validate_extractor(control_params: Dict[str, Any], timezone: str, **kwargs) -> None:
    delete_task_instances()
    ti = kwargs["ti"]
    success_tasks = ti.xcom_pull(task_ids="EXTRACTORS", key="SUCCESS_TASKS")
    failed_tasks = ti.xcom_pull(task_ids="EXTRACTORS", key="FAILED_TASKS")
    conf = ti.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="CONTROL-CONFIG")
    final_dict = {}
    status = ProcessStatusEnum.SUCCESS.value
    if not isinstance(success_tasks, type(None)) and len(success_tasks) > 0:
        for success_task in success_tasks:
            task = ti.xcom_pull(task_ids="EXTRACTORS", key=success_task)[0]
            final_dict.update({success_task: task})
    if not isinstance(failed_tasks, type(None)) and len(failed_tasks) > 0:
        status = ProcessStatusEnum.FAIL.value
        for failed_task in failed_tasks:
            task = ti.xcom_pull(task_ids="EXTRACTORS", key=failed_task)[0]
            final_dict.update({failed_task: task})
    conf = update_new_process(conf, status, final_dict, timezone)
    if status == ProcessStatusEnum.FAIL.value:
        conn = control_params["connection_id"]
        bucket = control_params["bucket"]
        prefix = control_params["prefix"]
        if not prefix.endswith("/"):
            prefix += "/"
        key = prefix + control_params["filename"]
        conf = json.dumps(conf, indent=2, default=str)
        loaded = load_obj_to_s3(bytes(conf.encode()), conn, bucket, key)
        if loaded:
            logger.info(f"Cargado correctamente el archivo de control en {key}")
        delete_all_xcom_tasks()
        raise AirflowSkipException(f"Ocurrieron errores en la etapa de extracción")
    elif status == ProcessStatusEnum.SUCCESS.value:
        ti.xcom_push(key="CONTROL-CONFIG", value=conf)


def on_failure_extractor(context) -> None:
    ti = context["ti"]
    task_name = f"{ti.task_id}_{ti.map_index}"
    selects = Variable.get('SELECTS', default_var=[], deserialize_json=True)
    command = selects[ti.map_index]
    tablename = select_multiple(command[1])["tablename"]
    exception = str(context["exception"])
    status = ProcessStatusEnum.FAIL.value
    task_result = {"description": tablename, "status": status, "message": exception}
    ti.xcom_push(key=task_name, value=task_result)
    ti.xcom_push(key="FAILED_TASKS", value=task_name)


def on_success_extractor(context) -> None:
    ti = context["ti"]
    task_name = f"{ti.task_id}_{ti.map_index}"
    selects = Variable.get('SELECTS', default_var=[], deserialize_json=True)
    command = selects[ti.map_index]
    tablename = select_multiple(command[1])["tablename"]
    status = ProcessStatusEnum.SUCCESS.value
    task_result = {"description": tablename, "status": status, "message": ""}
    ti.xcom_push(key=task_name, value=task_result)
    ti.xcom_push(key="SUCCESS_TASKS", value=task_name)


def extract_from_source(command: str, source_conn, intern_conn, chunksize: int, **kwargs):
    task = kwargs['ti']
    extract_type = command[0].split("|")[0]
    command = command[1]
    source_engine = source_conn.engine
    definitions = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="EXTRACTION-DEFINITION-JSON")
    logger.info(f"Definiciones desde json: {definitions}")
    multiple = select_multiple(command)
    tablename = multiple["tablename"]
    is_procedure = False
    if extract_type == "PROCEDURE":
        is_procedure = True
    columns_name = []
    # Create the model with the procedure descriptor
    model = None
    for procedure in definitions:
        if procedure["identifier"] != tablename:
            continue
        indexes = []
        if "indexes" in procedure.keys():
            indexes = procedure["indexes"]
        model = generateModel(tablename, procedure["fields"], indexes)
        columns_name = [field["name"] for field in procedure["fields"]]
    if isinstance(model, type(None)):
        raise AssertionError(f"Definición del extracción para {tablename} en el json-descriptor no encontraddo")
    try:
        create = intern_conn.create_table(model)
        if create:
            logger.info(f"Creado correctamente la tabla {tablename}. Creado?: {create}")
        else:
            raise AssertionError(f"Error creando tabla {tablename}")
        if is_procedure:
            command = command[len(tablename+"|"):]
            temp_connection = source_conn.get_basic_connection()
            cursor = temp_connection.cursor()
            final_command = source_conn.generate_sql_procedure(command)
            logger.debug(f"FINAL COMMAND: {final_command}")
            cursor.execute(final_command)
            for resultSet in cursor.getimplicitresults():
                data = []
                for row in resultSet:
                    data.append(row)
                    if len(data) == chunksize:
                        dataframe = pd.DataFrame(data, columns=columns_name)
                        save = save_from_dataframe(dataframe, tablename, intern_conn.engine)
                        if save:
                            logger.debug(f"Guardado correctamente dataframe. Procesando más bloques")
                        data.clear()
                if len(data) > 0:
                    dataframe = pd.DataFrame(data, columns=columns_name)
                    save = save_from_dataframe(dataframe, tablename, intern_conn.engine)
                    if save:
                        logger.debug(f"Migrado correctamente todos los datos")
                    data.clear()
            logger.info("Guardado correctamente todos los datos")
            source_conn.close_basic_connection()
        else:
            if command.replace(" ", "").lower().find("|select"):
                command = command[command.find("select"):]
            steps = get_steps(command, chunksize, source_engine)
            # Traemos el iterator
            iterator = get_iterator(command, chunksize, source_engine)
            logger.info(f"Número de pasos para migrar datos: {steps}")
            for step in range(steps):
                dataframe = next(iterator)
                dataframe["INTERN_ID_BCOM"] = None
                logger.debug(dataframe)
                save = save_from_dataframe(dataframe, tablename, intern_conn.engine)
                if save:
                    logger.info(f"Guardado correctamente dataframe en el paso {step+1}")
    except Exception as e:
        delete = delete_table(tablename, intern_conn.engine)
        if delete:
            logger.info(f"Se borró correctamente la tabla {tablename}")
        raise AssertionError(f"Error creando la tabla y migrando datos. {type(e)}. {e}")


@task(task_id="MASTER_EXTRACTOR", trigger_rule='all_success')
def get_select_from_xcom(**kwargs):
    task = kwargs['ti']
    final_selects = []
    conf = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="CONTROL-CONFIG")
    tasks = get_tasks_from_control(conf, "extractor")
    success_tasks = tasks["tasks"]
    success_tasks = [item[1] for item in success_tasks]
    logger.info(f"COMANDOS QUE FUERON EXITOSOS: {success_tasks}")
    xcom_keys = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="XCOM-EXTRACTION-NAMES")
    logger.debug(xcom_keys)
    for key in xcom_keys:
        if not key.startswith(OperationTypeEnum.SELECT.value) and not key.startswith(OperationTypeEnum.PROCEDURE.value):
            continue
        xcom_selects = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key=key)
        logger.info(f"Trayendo comandos {xcom_selects}")
        for select in xcom_selects:
            tablename = select_multiple(select)["tablename"]
            if tasks["status"] == ProcessStatusEnum.SUCCESS.value or tablename not in success_tasks:
                final_selects.append((key, select, ))
    logger.info(f"Comandos para la extracción: {final_selects}")
    Variable.set(key='SELECTS', value=final_selects, serialize_json=True)
    return [[item] for item in final_selects]


def get_extract_task_group(db_source_conn, db_intern_conn, chunksize: int, timezone: str,
                           control_s3: Dict[str, Any]) -> TaskGroup or None:
    group = None
    try:
        with TaskGroup(group_id="ExtraccionDeDatos", prefix_group_id=False) as group:
            selects = get_select_from_xcom()

            tasks = PythonOperator.partial(
                task_id="EXTRACTORS",
                python_callable=extract_from_source,
                op_kwargs={'source_conn': db_source_conn, 'intern_conn': db_intern_conn, 'chunksize': chunksize},
                on_failure_callback=on_failure_extractor,
                on_success_callback=on_success_extractor
            ).expand(op_args=selects)

            validate_task = PythonOperator(
                task_id="VALIDATE_EXTRACTION",
                python_callable=validate_extractor,
                op_kwargs={'control_params': control_s3, 'timezone': timezone},
                trigger_rule='all_done'
            )
            selects >> tasks >> validate_task
    except Exception as e:
        logger.error(f"Error creando taskGroup de extracción. {e}")
    finally:
        return group

