import json
from typing import Dict, Any
from airflow.utils.task_group import TaskGroup
from airflow.operators.python import PythonOperator
from airflow.models import Variable
from airflow.decorators import task
from airflow.exceptions import AirflowSkipException

from components.Control import get_tasks_from_control, update_new_process
from components.S3Route import load_obj_to_s3
from components.Xcom import delete_all_xcom_tasks, delete_task_instances
from enums.ProcessStatusEnum import ProcessStatusEnum
from enums.OperationTypeEnum import OperationTypeEnum

import logging

logger = logging.getLogger()


def validate_transform(control_params: Dict[str, Any], timezone: str, **kwargs) -> None:
    delete_task_instances()
    ti = kwargs["ti"]
    success_tasks = ti.xcom_pull(task_ids="TRANSFORMATIONS", key="SUCCESS_TASKS")
    failed_tasks = ti.xcom_pull(task_ids="TRANSFORMATIONS", key="FAILED_TASKS")
    conf = ti.xcom_pull(task_ids="VALIDATE_EXTRACTION", key="CONTROL-CONFIG")
    final_dict = {}
    status = ProcessStatusEnum.SUCCESS.value
    if not isinstance(success_tasks, type(None)) and len(success_tasks) > 0:
        for success_task in success_tasks:
            task = ti.xcom_pull(task_ids="TRANSFORMATIONS", key=success_task)[0]
            final_dict.update({success_task: task})
    if not isinstance(failed_tasks, type(None)) and len(failed_tasks) > 0:
        status = ProcessStatusEnum.FAIL.value
        for failed_task in failed_tasks:
            task = ti.xcom_pull(task_ids="TRANSFORMATIONS", key=failed_task)[0]
            final_dict.update({failed_task: task})
    conf = update_new_process(conf, status, final_dict, timezone, True)
    if status == ProcessStatusEnum.FAIL.value:
        conn = control_params["connection_id"]
        bucket = control_params["bucket"]
        prefix = control_params["prefix"]
        if not prefix.endswith("/"):
            prefix += "/"
        key = prefix + control_params["filename"]
        conf = json.dumps(conf, indent=2, default=str)
        loaded = load_obj_to_s3(bytes(conf.encode()), conn, bucket, key)
        if loaded:
            logger.info(f"Cargado correctamente el archivo de control en {key}")
        delete_all_xcom_tasks()
        raise AirflowSkipException(f"Ocurrieron errores en la etapa de transformación")
    elif status == ProcessStatusEnum.SUCCESS.value:
        ti.xcom_push(key="CONTROL-CONFIG", value=conf)


def on_failure_transform(context) -> None:
    ti = context["ti"]
    task_name = f"{ti.task_id}_{ti.map_index}"
    transform = Variable.get('TRANSFORMS', default_var=[], deserialize_json=True)
    script = transform[ti.map_index][0]
    exception = str(context["exception"])
    status = ProcessStatusEnum.FAIL.value
    task_result = {"description": script, "status": status, "message": exception}
    ti.xcom_push(key=task_name, value=task_result)
    ti.xcom_push(key="FAILED_TASKS", value=task_name)


def on_success_transform(context) -> None:
    ti = context["ti"]
    task_name = f"{ti.task_id}_{ti.map_index}"
    transform = Variable.get('TRANSFORMS', default_var=[], deserialize_json=True)
    script = transform[ti.map_index][0]
    status = ProcessStatusEnum.SUCCESS.value
    task_result = {"description": script, "status": status, "message": ""}
    ti.xcom_push(key=task_name, value=task_result)
    ti.xcom_push(key="SUCCESS_TASKS", value=task_name)


def transformations(xcom_commands: str, intern_conn):
    engine = intern_conn.engine
    script_name = xcom_commands[0]
    commands = xcom_commands[1]
    logger.info(f"Ejecutando transformaciones del script {script_name}")
    with engine.connect() as connection:
        for command in commands:
            logger.info(f"Ejecutando comando de transformación: {command}")
            _ = connection.execute(command)


@task(task_id="MASTER_TRANSFORMATION", trigger_rule='none_skipped')
def get_trans_from_xcom(**kwargs):
    task = kwargs['ti']
    transforms_per_file = []
    conf = task.xcom_pull(task_ids="VALIDATE_EXTRACTION", key="CONTROL-CONFIG")
    tasks = get_tasks_from_control(conf, "transformation")
    success_tasks = tasks["tasks"]
    success_tasks = [item[1] for item in success_tasks]
    logger.info(f"SCRIPTS QUE FUERON EXITOSOS: {success_tasks}")
    xcom_keys = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="XCOM-EXTRACTION-NAMES")
    logger.debug(xcom_keys)
    for key in xcom_keys:
        final_transforms = []
        if not key.startswith(OperationTypeEnum.TRANSFORM.value):
            continue
        xcom_transforms = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key=key)
        logger.info(f"Trayendo comandos {xcom_transforms}")
        if tasks["status"] == ProcessStatusEnum.SUCCESS.value or key not in success_tasks:
            for transform in xcom_transforms:
                final_transforms.append(transform)
            transforms_per_file.append((key, final_transforms))
    logger.info(f"Scripts para la transformación: {transforms_per_file}")
    Variable.set(key='TRANSFORMS', value=transforms_per_file, serialize_json=True)
    return [[item] for item in transforms_per_file]


def get_transform_task_group(db_intern_conn, timezone: str, control_s3: Dict[str, Any]) -> TaskGroup or None:
    group = None
    try:
        with TaskGroup(group_id="TransformacionDeDatos", prefix_group_id=False) as group:
            transforms = get_trans_from_xcom()

            tasks = PythonOperator.partial(
                task_id="TRANSFORMATIONS",
                python_callable=transformations,
                op_kwargs={'intern_conn': db_intern_conn},
                on_failure_callback=on_failure_transform,
                on_success_callback=on_success_transform
            ).expand(op_args=transforms)

            validate_task = PythonOperator(
                task_id="VALIDATE_TRANSFORMATION",
                python_callable=validate_transform,
                op_kwargs={'control_params': control_s3, 'timezone': timezone},
                trigger_rule='none_skipped'
            )
            transforms >> tasks >> validate_task
    except Exception as e:
        logger.error(f"Error creando taskGroup de transformación. {e}")
    finally:
        return group
