from datetime import datetime, timedelta
from typing import Any, Dict

import time
import pandas as pd
import numpy as np

from components.S3Route import get_df_from_s3, get_base_date, save_df_to_s3, move_object_s3
from components.Sensor import create_s3_sensor
from components.Utils import get_modified_prefix, remove_invalid_rows, remove_fields, update_dict_with_catalogs

from airflow import DAG
from airflow.operators.python import PythonOperator

import logging

logger = logging.getLogger()

DAG_NAME = "BCOM_DAG_TRANSFORMACION_TACOMVENTAS_PROMOCIONESRESIDENCIAL"
PROMOCION_3PA2P = "3P a 2P"
PROMOCION_PARRILLA = "Parrilla"
PROMOCION_HBO = "Promocion HBO"
PROMOCION_DEFAULT = "Promocion"
PROMOCIONES_NO_CONSIDERADAS_TV_CANALES = "Adicional|Soundbox|SOUNDBOX"

DEFAULT_ARGS = {
    'owner': 'BCOM',
    "start_date": datetime(2023, 5, 25, 22, 9),
    'depends_on_past': False,
    'email': 'caguirre@bytesw.com',
    'retries': 1,
    'retries_delay': timedelta(minutes=3),
    'email_on_retry': False,
    'email_on_failure': False
}


def dag1_id1(promo: pd.DataFrame, catalog_promo: pd.DataFrame, key_field: str, value_field: str) -> pd.DataFrame:
    try:
        catalog = catalog_promo.values
        for item in catalog:
            promo[value_field] = np.where(promo[key_field].str.contains(item[0]), item[1], promo[value_field])
    except Exception as e:
        logger.error(f"Error DAG1_ID1. {e}")
    return promo


def dag1_id2(promo: pd.DataFrame, relation_poid: pd.DataFrame, key_field: str, value_field: str) -> pd.DataFrame:
    try:
        relation_poid = relation_poid.rename(columns={value_field: 'NUEVO_'+value_field})
        relation_poid = relation_poid.drop_duplicates(subset=[key_field])
        promo = promo.merge(relation_poid, 'left', key_field)
        promo[value_field] = np.where(promo['NUEVO_'+value_field].notna(), promo['NUEVO_'+value_field], promo[value_field])
        promo = promo.drop('NUEVO_'+value_field, axis=1)
    except Exception as e:
        logger.error(f"Error DAG1_ID2. {e}")
    return promo


def dag1_id3(tacom: pd.DataFrame, promo: pd.DataFrame) -> pd.DataFrame:
    result = pd.DataFrame()
    try:
        promo["CD_PAQUETE_PROMO"] = promo["CD_PAQUETE"].astype(int, errors='ignore')
        promo.drop("CD_PAQUETE", axis=1, inplace=True)
        promo = promo.drop_duplicates(["CUENTA", "CD_PAQUETE_PROMO"])
        result = tacom.merge(promo, how='left', left_on=["CD_CUENTA", "CD_PAQUETE"], right_on=["CUENTA", "CD_PAQUETE_PROMO"])
        result["CD_PAQUETE"] = result["CD_PAQUETE"].astype(int).astype(str)

        result["PROMOCION"] = np.where((result["CD_PAQUETE_PROMO"].isna()) | (result["CD_PAQUETE_PROMO"] == "None") |
                                       (result["CD_PAQUETE_PROMO"] == "nan"), None,
                              np.where((result["CD_PAQUETE"].notna()) & (result["CD_PAQUETE"].str.len() <= 5),
                                       PROMOCION_3PA2P,
                              np.where((result["NOMBRE_PRODUCTO"].str.contains("TV", na=False)) &
                                       (~result["NOMBRE_PRODUCTO"].str.contains(PROMOCIONES_NO_CONSIDERADAS_TV_CANALES, na=False)),
                                       PROMOCION_PARRILLA,
                                       np.where((result["NOMBRE_PRODUCTO"].str.contains("CANALES", na=False)) &
                                                (~result["NOMBRE_PRODUCTO"].str.contains(PROMOCIONES_NO_CONSIDERADAS_TV_CANALES, na=False)),
                                                PROMOCION_PARRILLA,
                                                np.where(result["NOMBRE_PRODUCTO"].str.contains("HBO MAX", na=False), PROMOCION_HBO,
                              np.where(result["NOMBRE_PRODUCTO"].str.contains("PAQUETE HBO", na=False), PROMOCION_HBO,
                              np.where(result["NOMBRE_PRODUCTO"].str.contains("STAR PREMIUM", na=False),
                                       "Promocion STAR PREMIUM", PROMOCION_DEFAULT)))))))

        result["CD_PAQUETE_PROMO"] = np.where((result["CD_PAQUETE_PROMO"] == 'nan') |
                                              (result["CD_PAQUETE_PROMO"] == 'None'), None, result["CD_PAQUETE"])
    except Exception as e:
        logger.error(f"Error DAG1_ID3. {e}")
    finally:
        return result


def dag1_id4(df: pd.DataFrame, df_promo: pd.DataFrame, key_field: str, value_field: str) -> pd.DataFrame:
    try:
        df["CD_PAQUETE"] = df["CD_PAQUETE"].astype(str, errors='ignore')
        df_promo[key_field] = df_promo[key_field].astype(str, errors='ignore')
        df_promo[value_field] = df_promo[value_field].astype(str, errors='ignore')
        df = df.merge(df_promo, how='outer', left_on="CD_PAQUETE", right_on=key_field)
        df = df.dropna(how='all', subset=["CD_EMPRESA", "CD_FOLIO", "CD_CUENTA"])
        df["CD_PAQUETE"] = np.where((df["PROMOCION"] == PROMOCION_3PA2P) & (df[key_field].notna()), df[value_field], df["CD_PAQUETE"])
        df = df.drop([key_field, value_field], axis=1)
    except Exception as e:
        logger.error(f"Error DAG1_ID4. {e}")
    finally:
        return df


def dag1_id5(df: pd.DataFrame, relation: pd.DataFrame, key_field: str, value_field: str) -> pd.DataFrame:
    try:
        relation = relation.astype(str)
        df = df.merge(relation, how='left', left_on="CD_PAQUETE", right_on=key_field)
        df["CD_PAQUETE"] = np.where(df[value_field].notna(), df[value_field], df["CD_PAQUETE"])
    except Exception as e:
        logger.error(f"Error DAG1_ID5. {e}")
    finally:
        return df


def dag1_id6(df: pd.DataFrame) -> pd.DataFrame:
    try:
        df["NRO_PAQUETE"] = df.groupby(["CD_FOLIO", "CD_PAQUETE"]).cumcount() + 1
    except Exception as e:
        logger.error(f"Error DAG1_ID6. {e}")
    finally:
        return df


def dag1_id7(df: pd.DataFrame, notpromo: pd.DataFrame, key_field: str) -> pd.DataFrame:
    try:
        notpromo = notpromo.astype(str)
        not_promo_values = notpromo[key_field].tolist()
        df["PROMOCION"] = np.where((df["TP_SERVICIO"] == 2) & (df["NRO_PAQUETE"] > 1) &
                                   (df["CD_PAQUETE"].isin(not_promo_values)), None, df["PROMOCION"])
    except Exception as e:
        logger.error(f"Error DAG1_ID7. {e}")
    finally:
        return df


def etl_dag1(conn: str, bucket: str, tacom_key: str, promo_key: str, catalogs: Dict[str, Any],
             delimiter: str, interval: str, outputs: Dict[str, Any], procesed_prefix: str, filters: Dict[str, Any]):
    start_time = time.time()
    logger.debug(f"DICCIONARIO DE CATALOGOS: {catalogs}")
    base_date = get_base_date(conn, bucket, tacom_key)
    omited_cols = filters["fields_omited"]

    current_delimiter = delimiter
    tacom = get_df_from_s3(conn, bucket, tacom_key, current_delimiter, base_date, interval)
    tacom_df = remove_invalid_rows(tacom['df'], filters["tacom_drop_nulls_subset"])
    tacom_df = remove_fields(tacom_df, omited_cols)
    if isinstance(tacom_df, type(None)):
        raise AssertionError(f"INSUMO TACOM NO ENCONTRADO")
    promo = get_df_from_s3(conn, bucket, promo_key, current_delimiter, base_date, interval)
    promo_df = remove_invalid_rows(promo['df'], filters["promo_drop_nulls_subset"])
    promo_df = remove_fields(promo_df, omited_cols)
    if isinstance(promo_df, type(None)):
        raise AssertionError(f"INSUMO PROMOCIONES NO ENCONTRADO")

    if "catalogo_promociones_delimiter" in catalogs.keys():
        current_delimiter = catalogs["catalogo_promociones_delimiter"]
    promo_catalog = get_df_from_s3(conn, bucket, catalogs["s3_catalogo_promociones"], current_delimiter, base_date, interval)
    promo_catalog_df = remove_invalid_rows(promo_catalog['df'], filters["catalog_drop_nulls_subset"])
    promo_catalog_df = remove_fields(promo_catalog_df, omited_cols)
    if isinstance(promo_catalog_df, type(None)):
        raise AssertionError(f"INSUMO PROMOCIONES CATALOGO NO ENCONTRADO")

    if "relacionpoidpaquete_delimiter" in catalogs.keys():
        current_delimiter = catalogs["relacionpoidpaquete_delimiter"]
    relationpoid = get_df_from_s3(conn, bucket, catalogs["s3_relacionpoidpaquete"], current_delimiter, base_date, interval)
    relationpoid_df = remove_invalid_rows(relationpoid['df'], filters["relapoid_drop_nulls_subset"])
    relationpoid_df = remove_fields(relationpoid_df, omited_cols)
    if isinstance(relationpoid_df, type(None)):
        raise AssertionError(f"INSUMO RELACION POID NO ENCONTRADO")

    if "relacion3pa2p_delimiter" in catalogs.keys():
        current_delimiter = catalogs["relacion3pa2p_delimiter"]
    relation3a2p = get_df_from_s3(conn, bucket, catalogs["s3_relacion3pa2p"], current_delimiter, base_date, interval)
    relation3a2p_df = remove_invalid_rows(relation3a2p['df'], filters["rela3pa2p_drop_nulls_subset"])
    relation3a2p_df = remove_fields(relation3a2p_df, omited_cols)
    if isinstance(relation3a2p_df, type(None)):
        raise AssertionError(f"INSUMO RELACION 3A2P NO ENCONTRADO")

    if "relacion_paquetes_delimiter" in catalogs.keys():
        current_delimiter = catalogs["relacion_paquetes_delimiter"]
    relapaq_catalog = get_df_from_s3(conn, bucket, catalogs["s3_relacion_paquetes"], current_delimiter, base_date, interval)
    relapaq_df = remove_invalid_rows(relapaq_catalog['df'], filters["relapaqs_drop_nulls_subset"])
    relapaq_df = remove_fields(relapaq_df, omited_cols)
    if isinstance(relapaq_df, type(None)):
        raise AssertionError(f"INSUMO RELACION PAQUETE INICIAL Y FINAL NO ENCONTRADO")

    if "no_promocion_delimiter" in catalogs.keys():
        current_delimiter = catalogs["no_promocion_delimiter"]
    not_promo_catalog = get_df_from_s3(conn, bucket, catalogs["s3_no_promocion"], current_delimiter, base_date, interval)
    not_promo_df = remove_invalid_rows(not_promo_catalog['df'], filters["not_promo_drop_nulls_subset"])
    not_promo_df = remove_fields(not_promo_df, omited_cols)
    if isinstance(not_promo_df, type(None)):
        raise AssertionError(f"INSUMO PAQUETES SIN PROMOCION NO ENCONTRADO")

    logger.info(f"EJECUTANDO PROCESO")
    promo_df = dag1_id1(promo_df, promo_catalog_df, catalogs["catalogo_promociones_key"],
                        catalogs["catalogo_promociones_value"])
    promo_df = dag1_id2(promo_df, relationpoid_df, catalogs["relacionpoidpaquete_key"],
                        catalogs["relacionpoidpaquete_value"])
    result = dag1_id3(tacom_df, promo_df)
    result = dag1_id4(result, relation3a2p_df, catalogs["relacion3pa2p_key"], catalogs["relacion3pa2p_value"])
    result = dag1_id5(result, relapaq_df, catalogs["relacion_paquetes_key"], catalogs["relacion_paquetes_value"])
    result = dag1_id6(result)
    result = dag1_id7(result, not_promo_df, catalogs["no_promocion_key"])
    final_columns = list(tacom_df.columns)
    result = result[final_columns]
    result = result.sort_values(by="CD_FOLIO")

    # Subir los resultados al S3
    save_df_to_s3(result, conn, bucket, outputs["tacom_output_path"], outputs["tacom_delimiter"])
    promo_df.rename(columns={'CD_PAQUETE_PROMO': 'CD_PAQUETE'}, inplace=True)
    save_df_to_s3(promo_df, conn, bucket, outputs["promo_output_path"], outputs["promo_delimiter"])

    # Mover TODOS LOS INSUMOS USADOS
    move_object_s3(conn, bucket, tacom['filename'], procesed_prefix)
    move_object_s3(conn, bucket, promo['filename'], procesed_prefix)
    move_object_s3(conn, bucket, promo_catalog['filename'], procesed_prefix)
    move_object_s3(conn, bucket, relationpoid['filename'], procesed_prefix)
    move_object_s3(conn, bucket, relation3a2p['filename'], procesed_prefix)
    move_object_s3(conn, bucket, relapaq_catalog['filename'], procesed_prefix)
    move_object_s3(conn, bucket, not_promo_catalog['filename'], procesed_prefix)

    logger.info("FINALIZADO PROCESO 1")
    logger.info(f"--- {time.time() - start_time} seconds ---  TIEMPO DE EJECUCIÓN DEL PROCESO 1")


def set_dag_1():
    """ DAG that execute the process with TACOMVENTAS AND PROMOCIONES_RESIDENCIAL inputs"""

    import yaml
    from yaml.loader import SafeLoader
    # Cambiar conf_path dependiendo del ambiente, en prod usando k8 y contenedores usar /opt/airflow/dags/app_conf.yml
    # En desarrollo, cualquiera que apunte a su carpeta dags
    conf_path = "/opt/airflow/dags/app_conf.yml"
    with open(conf_path) as f:
        data = yaml.load(f, Loader=SafeLoader)
        general_cnf = data["general"]
        conf = data["dags"]["dag1"]
    with DAG(DAG_NAME, default_args=DEFAULT_ARGS, description="Proceso que extrae y transforma",
             schedule_interval=conf["schedule"], tags=["DAG 1"], catchup=True) as dag:
        dag1_inputs_s3_params = conf["s3_parameters"]["inputs"]
        s3_conf = general_cnf["s3_parameters"]

        s3_prefix = get_modified_prefix(dag1_inputs_s3_params["prefix"])
        s3_tacom = s3_prefix + dag1_inputs_s3_params["tacom_pattern"]
        s3_promo = s3_prefix + dag1_inputs_s3_params["promociones_pattern"]

        catalogs_dict = {}
        catalogs_dict = update_dict_with_catalogs(catalogs_dict, conf, "catalogo_promociones", s3_prefix)
        catalogs_dict = update_dict_with_catalogs(catalogs_dict, conf, "relacion3pa2p", s3_prefix)
        catalogs_dict = update_dict_with_catalogs(catalogs_dict, conf, "relacionpoidpaquete", s3_prefix)
        catalogs_dict = update_dict_with_catalogs(catalogs_dict, conf, "relacion_paquetes", s3_prefix)
        catalogs_dict = update_dict_with_catalogs(catalogs_dict, conf, "no_promocion", s3_prefix)

        # Define the sensor to verify if data exists or have been updated
        s3_sensor_tacom = create_s3_sensor("S3_sensor_tacom_task", s3_conf["s3_conn_id"], s3_conf["bucket"],
                                           s3_tacom)

        s3_sensor_promo = create_s3_sensor("S3_sensor_promo_task", s3_conf["s3_conn_id"], s3_conf["bucket"],
                                           s3_promo)

        s3_sensor_promo_catalog = create_s3_sensor("S3_sensor_promo_catalog_task", s3_conf["s3_conn_id"],
                                                   s3_conf["bucket"], catalogs_dict["s3_catalogo_promociones"])

        s3_sensor_3a2p = create_s3_sensor("S3_sensor_3a2p_task", s3_conf["s3_conn_id"], s3_conf["bucket"],
                                          catalogs_dict["s3_relacion3pa2p"])

        s3_sensor_poid = create_s3_sensor("S3_sensor_poid_task", s3_conf["s3_conn_id"], s3_conf["bucket"],
                                          catalogs_dict["s3_relacionpoidpaquete"])

        s3_sensor_paq = create_s3_sensor("S3_sensor_paq_task", s3_conf["s3_conn_id"], s3_conf["bucket"],
                                         catalogs_dict["s3_relacion_paquetes"])

        s3_sensor_notpromo = create_s3_sensor("S3_sensor_notpromo_task", s3_conf["s3_conn_id"], s3_conf["bucket"],
                                              catalogs_dict["s3_no_promocion"])

        outputs = conf["s3_parameters"]["outputs"]
        output_prefix = outputs["prefix"]
        output_prefix = get_modified_prefix(output_prefix)
        output_tacom_path = output_prefix + outputs["tacom_output"]
        output_tacom_sep = ","
        if "tacom_delimiter" in outputs.keys():
            output_tacom_sep = outputs["tacom_delimiter"]
        output_promo_path = output_prefix + outputs["promo_output"]
        output_promo_sep = ","
        if "promo_delimiter" in outputs.keys():
            output_promo_sep = outputs["promo_delimiter"]
        outputs = {'tacom_output_path': output_tacom_path, 'tacom_delimiter': output_tacom_sep,
                   'promo_output_path': output_promo_path, 'promo_delimiter': output_promo_sep}
        procesed_prefix = conf["s3_parameters"]["procesed_prefix"]
        if not procesed_prefix.endswith("/") and procesed_prefix != "":
            procesed_prefix += "/"

        filters = conf["filters"]
        transformation = PythonOperator(
            task_id="ETL-DAG1",
            python_callable=etl_dag1,
            op_kwargs={'conn': s3_conf["s3_conn_id"], 'bucket': s3_conf["bucket"],
                       'tacom_key': s3_tacom, 'promo_key': s3_promo, 'catalogs': catalogs_dict,
                       'delimiter': conf["csv_delimiter"], 'interval': conf["group_input_interval_days"],
                       'outputs': outputs, 'procesed_prefix': procesed_prefix, 'filters': filters},
            trigger_rule="all_success"
        )

        (s3_sensor_tacom, s3_sensor_promo, s3_sensor_promo_catalog, s3_sensor_3a2p,
         s3_sensor_poid, s3_sensor_paq, s3_sensor_notpromo) >> transformation
    return dag


globals()["0"] = set_dag_1()

